package ac.essex.ooechs.adaboost;

import java.io.Serializable;
import java.text.DecimalFormat;

/**
 * Implementation of Freud and Schapire's Adaboost M1 Algorithm which extends their problem
 * to boost multiclass classifiers. Unlike the original Adaboost algorithm which expects each
 * sample to be {true, false}, in AdaboostM1 each sample is allowed any truth label {1,...,k}
 * so multiclass problems can be addressed.
 *
 * To use AdaBoost, you must override this abstract class and implement the getSamples() method
 * and the weakLearn() function, inserting whatever learning algorithm you want. You can then call
 * the boost() method to start the learning process. Your weakLearn function must return an object
 * which extends the AdaBoost learner object, that is, be able to do something to an AdaBoost sample
 * and return a classification.
 *
 * @author Olly Oechsle, University of Essex, Date: 03-May-2007
 * @version 1.0
 */
public abstract class AdaBoost implements Serializable {

    /**
     * Returns an array of multiclass sample objects which make up the training data
     */
    public abstract AdaBoostSample[] getSamples();

    /**
     * Runs whatever learning algorithm it is to find a rough solution.
     * @param samples The training samples on which to teach the learner
     * @param d The weights of each sample which may be used as fitness cues for the learner
     */
    protected abstract AdaBoostLearner weakLearn(AdaBoostSample[] samples, double[] d);

    /**
     * After AdaBoosting, we are left with a set of weak learners
     */
    protected AdaBoostLearner[] h;

    /**
     * With each weak learner is an associated beta value which is a measure of the
     * learner's accuracy.
     */
    protected double b[];        

    /**
     * Runs the adaboost algorithm
     * @param T The number of boosting stages to do.
     */
    public void boost(final int T) {

        // get the training samples
        AdaBoostSample[] samples = getSamples();

        // m is the number of training data items
        int m = samples.length;

        // distribution d are the weights for each sample
        double[] d = new double[m];

        // initialise d, with all samples equally weighted
        DecimalFormat f = new DecimalFormat("0.00000000");
        final double average = 1 / (double) m;
        for (int i = 0; i < d.length; i++) {
            d[i] = average;
        }

        // the hypotheses of each iteration are saved here
        h = new AdaBoostLearner[T];

        // values associated with each learner which indicate how accurate the learner is
        // and therefore how confident we should be in its decision.
        b = new double[T];

        // Do for t = 1,2,...,T
        for (int t = 0; t < T; t++) {

            message("ADABOOST: t=" + t);

            // Call weak learn, providing it with the distribution Dt
            h[t] = weakLearn(samples, d);

            // Calculate the error of ht which is the sum of weights for which ht was incorrect
            double e = 0;
            boolean[] correct = new boolean[m];
            for (int i = 0; i < samples.length; i++) {
                correct[i] = h[t].classify(samples[i], null) == samples[i].label;
                if (!correct[i]) e += d[i];
                // record the success of individual classifiers for reference
                else h[t].individualTP++;
            }

            // individual is not good enough
            if (e > 0.5) {
                // remove this classifier
                message("ADABOOST: Rejecting individual: e " + e);
                h[t] = null;
                break;
            }

            // no errors - found ideal individual
            if (e == 0 ) {
                message("ADABOOST: Ideal Individual");
                break;
            }

            // set the beta value associated with this hypothesis
            // using an exponential function that gives an exponential
            // distribution from 0 -> 1 for values 0 -> 0.5
            b[t] = e / (1 - e);

            // update the distribution. The lower the error, the lower b will be, so if instance i
            // was classified correctly it is multiplied by b, which reduces the weight of that sample.
            // If a sample is measured incorrectly, its weight is not changed.
            double dTotal = 0;
            for (int i = 0; i < samples.length; i++) {
                d[i] =  d[i] * (correct[i] ? b[t] : 1);
                dTotal += d[i];
            }

            // normalise the distribution so that all the weights add up to 1.0
            for (int i = 0; i < samples.length; i++) {
                d[i] /= dTotal;
            }

            if (test(samples) == samples.length) {
                message("Found ideal classifier.");
                break;
            }

            onIterationEnd();

        }

        finish();

    }

    public void onIterationEnd() {
        // do nothing
    }

    public void finish() {
        message("AdaBoost complete");
    }

    public void message(String message) {
        System.out.println(message);
    }

    public int classify(AdaBoostSample sample) {

        // each weak learner is run, whatever label that learner finds is given a value
        // which is the log of the learner's beta value, so learners with low error rates will
        // return a higher value. Whichever label has the highest score is the one that gets returned.

        // create an array that can accommodate 100 classIDs
        double[] classSums = new double[100];

        // the label to be outputted
        int mostLikelyLabel = -1;

        // the highest score for any label
        double highestScore = Double.MIN_VALUE;

        for (int t = 0; t < h.length; t++) {

            // in case h does not contain T classifiers
            if (h[t] == null) continue;

            // Get the learner's output: a class label
            int label = h[t].classify(sample, null);

            // Add a value to the sum associated with this label
            classSums[label] += Math.log10(1 / b[t]);

            // if this label has the highest score,
            // set this label as the one to be returned
            if (classSums[label] > highestScore) {
                highestScore = classSums[label];
                mostLikelyLabel = label;
            }

        }

        // return the label that has the highest score
        return mostLikelyLabel;

    }


    /**
     * Tests the ADABoost on a set of samples
     */
    public int test(AdaBoostSample[] samples) {

        int adaboostTP = 0;

        for (int i = 0; i < samples.length; i++) {
            AdaBoostSample sample = samples[i];
            int received = classify(sample);
            if (received == sample.label) adaboostTP++;
        }

        double percentage = (adaboostTP / (double) samples.length);
        DecimalFormat f = new DecimalFormat("0.000");

        message("Adaboost TP: " + adaboostTP + " / " + samples.length + "( " + f.format(percentage * 100) + "%, e=" + f.format(1-percentage) + ")");

        return adaboostTP;

    }

    /**
     * Returns the solution as an AdaBoostSolution object, which is a lightweight, serializable
     * and standalone class which can be used to deploy whatever solution you have.
     */
    public AdaBoostSolution getSolution() {

        AdaBoostSolution solution = new AdaBoostSolution(h.length);

        for (int i = 0; i < h.length; i++) {
            solution.addHypothesis(h[i], b[i]);
        }

        return solution;

    }

}
